from .filters import bill_webhook_filter, transaction_webhook_filter
from .logger import InterceptHandler
from .telegram import TelegramPollingProxy, TelegramWebhookProxy, BaseProxy

__all__ = (
    'TelegramPollingProxy',
    'TelegramWebhookProxy',
    'BaseProxy',
    'bill_webhook_filter',
    'transaction_webhook_filter',
    'InterceptHandler'
)



from typing import Optional, Union

from glQiwiApi.types.qiwi_types import currency_parsed as cur


class Currency:
    """
    Class with many currencies
    >>> usd = Currency.get('840')
    >>> usd
    ... cur.CurrencyModel(code='USD', decimal_digits=2, name='US Dollar', name_plural='US dollars',
    ...                   rounding=0, symbol='$', symbol_native='$')

    >>> usd.symbol
    ... '$'
    """

    @classmethod
    def get(cls, currency_code: Union[str, int]) -> Optional[
        cur.CurrencyModel
    ]:
        """
        Implements class-based getitem behaviour

        >>> Currency.get('840').symbol
        ... '$'
        >>> Currency.get('USD').symbol
        ... '$'
        :param currency_code: ISO 4217 string or CODE
        :return: Currency object
        """
        try:
            if isinstance(currency_code, int) or currency_code.isdigit():
                return cur.described.get(cur.codes_number[str(currency_code)])
            else:
                return cur.described.get(currency_code.upper())
        except (KeyError, AssertionError):
            raise RuntimeError(
                f"Currency code `{currency_code}` was not found"
            )



from pydantic import Field

from glQiwiApi.types.base import Base
from glQiwiApi.types.basics import Sum


class Commission(Base):
    """
    Комиссия за платеж

    """
    provider_id: int = Field(alias="providerId")
    withdraw_sum: Sum = Field(alias="withdrawSum")
    enrollment_sum: Sum = Field(alias="enrollmentSum")
    qiwi_commission: Sum = Field(alias="qwCommission")
    withdraw_to_enrollment_rate: int = Field(alias="withdrawToEnrollmentRate")


from dataclasses import dataclass
from typing import Union, Any, Optional

from pydantic import BaseModel, validator

DEFAULT_CACHE_TIME = 0


class Sum(BaseModel):
    """
    Сумма платежа

    """
    amount: Union[int, float, str]
    currency: Any

    @validator("currency", pre=True)
    def humanize_pay_currency(cls, v):
        from glQiwiApi.utils.currency_util import Currency

        if not isinstance(v, int):
            try:
                v = int(v)
            except ValueError:
                return v
        return Currency.get(str(v))


class OptionalSum(BaseModel):
    """ object: OptionalSum """
    value: Union[int, float]
    currency: str


class Type(BaseModel):
    """
    Базовая модель типа данных

    """
    id: str
    title: str


@dataclass
class Attributes:
    """
    Аттрибуты кэшированного запроса

    """
    headers: Optional[dict] = None
    json: Optional[dict] = None
    params: Optional[dict] = None
    data: Optional[dict] = None

    @classmethod
    def format(cls, kwargs: dict, args: tuple):
        return cls(
            **{k: kwargs.get(k) for k in args if
               isinstance(kwargs.get(k), dict)}
        )


@dataclass
class Cached:
    """
    Объект кэшированного запроса

    """
    kwargs: Attributes
    response_data: Any
    method: Optional[Any]
    status_code: Union[str, int, None] = None


__all__ = [
    'Sum',
    'OptionalSum',
    'Type',
    'Cached',
    'DEFAULT_CACHE_TIME',
    'Attributes'
]



"""Main model: Partner"""
from typing import List, Optional

from glQiwiApi.types.base import Base


class Partner(Base):
    """ Base partner class """
    title: str
    id: int

    maps: Optional[List[str]] = None


__all__ = ("Partner",)



from __future__ import annotations

import contextvars
import copy
from typing import Any, TYPE_CHECKING, TypeVar, Optional, cast, Generic, ClassVar, Dict

if TYPE_CHECKING:
    from glQiwiApi.core.aiohttp_custom_api import RequestManager


class ToolsMixin(object):
    """ Object: ToolsMixin """
    _requests: RequestManager

    async def __aenter__(self):
        """Создаем сессию, чтобы не пересоздавать её много раз"""
        await self._requests.create_session()
        return self

    async def close(self):
        """ shutdown wrapper, close aiohttp session and clear storage """
        await self._requests.close()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Закрываем сессию и очищаем кэш при выходе"""
        await self.close()

    def _get(self, item: Any) -> Any:
        try:
            return super().__getattribute__(item)
        except AttributeError:
            return None

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __deepcopy__(self, memo):
        cls = self.__class__
        kw: Dict[str, bool] = {"__copy_signal__": True}
        result = cls.__new__(cls, **kw)
        memo[id(self)] = result
        dct = {
            slot: self._get(slot) for slot in self.__slots__ if
            self._get(slot) is not None
        }
        for k, value in dct.items():
            if k == '_requests':
                value._session = None
            elif k == 'dispatcher':
                value._loop = None
            setattr(result, k, copy.deepcopy(value, memo))
        return result

    @property
    def data(self):
        data = getattr(self, '_data', None)
        if data is None:
            data = {}
            setattr(self, '_data', data)
        return data

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __contains__(self, key):
        return key in self.data

    def get(self, key, default=None):
        return self.data.get(key, default)


ContextInstance = TypeVar("ContextInstance")


class ContextInstanceMixin(Generic[ContextInstance]):
    __context_instance: ClassVar[contextvars.ContextVar[ContextInstance]]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__()
        cls.__context_instance = contextvars.ContextVar(f"instance_{cls.__name__}")

    @classmethod  # noqa: F811
    def get_current(  # noqa: F811
            cls, no_error: bool = True
    ) -> Optional[ContextInstance]:  # pragma: no cover  # noqa: F811
        """ Get current instance from context """
        # on mypy 0.770 I catch that contextvars.ContextVar always contextvars.ContextVar[Any]
        cls.__context_instance = cast(
            contextvars.ContextVar[ContextInstance], cls.__context_instance
        )

        try:
            current: Optional[ContextInstance] = cls.__context_instance.get()
        except LookupError:
            if no_error:
                current = None
            else:
                raise

        return current

    @classmethod
    def set_current(cls, value: ContextInstance) -> contextvars.Token[ContextInstance]:
        if not isinstance(value, cls):
            raise TypeError(
                f"Value should be instance of {cls.__name__!r} not {type(value).__name__!r}"
            )
        return cls.__context_instance.set(value)

    @classmethod
    def reset_current(cls, token: contextvars.Token[ContextInstance]) -> None:
        cls.__context_instance.reset(token)



import datetime
from typing import Optional, List

from pydantic import Field

from glQiwiApi.types import Sum
from glQiwiApi.types.base import Base


class OrderDetails(Base):
    """ object: OrderDetails """
    order_id: str = Field(..., alias="id")
    card_alias: str = Field(..., alias="cardAlias")
    status: str
    price: Optional[Sum] = None
    card_id: Optional[str] = Field(alias="cardId", default=None)


class CardCredentials(Base):
    """object: CardCredentials"""
    qvx_id: int = Field(..., alias="id")
    masked_pan: str = Field(..., alias="maskedPan")
    status: str
    card_expire: datetime.datetime = Field(..., alias="cardExpire")
    card_type: str = Field(..., alias="cardType")
    card_alias: Optional[str] = Field(..., alias="cardAlias")
    activated: datetime.datetime
    sms_recender: Optional[datetime.datetime] = Field(..., alias="smsResended")
    post_number: Optional[str] = Field(default=None, alias="postNumber")
    blocked_date: Optional[datetime.datetime] = Field(default=None,
                                                      alias="blockedDate")
    card_id: int = Field(..., alias="cardId")
    txn_id: int = Field(..., alias="txnId")
    card_expire_month: str = Field(..., alias="cardExpireMonth")
    card_expire_year: str = Field(..., alias="cardExpireYear")


class Requisite(Base):
    """object: Requisite"""
    name: str
    value: str


class Details(Base):
    """object: Details"""
    info: str
    description: str
    tariff_link: str = Field(..., alias="tariffLink")
    offer_link: str = Field(..., alias="offerLink")
    features: List[str]
    requisites: List[Requisite]


class CardInfo(Base):
    """object: CardInfo"""
    id_: int = Field(..., alias="id")
    name: str
    alias: str
    price: Sum
    period: str
    type_: str = Field(..., alias="type")
    details: Details


class Card(Base):
    """
    object: Card
    description: Данные выпущенных карт
    """
    details: CardCredentials = Field(..., alias="qvx")
    balance: Optional[Sum] = None
    info: CardInfo


__all__ = (
    'OrderDetails',
    'Card'
)



from __future__ import annotations

import abc
import asyncio
import typing
from asyncio import AbstractEventLoop

from aiohttp import web

if typing.TYPE_CHECKING:
    try:
        from aiogram import Dispatcher  # type: ignore
    except (ModuleNotFoundError, ImportError):  # type: ignore
        pass

__all__ = ["TelegramWebhookProxy", "TelegramPollingProxy", "BaseProxy"]

# Some aliases =)
ListOfRoutes = typing.List[web.ResourceRoute]
SubApps = typing.List[typing.Tuple[str, web.Application, ListOfRoutes]]


def _init_sub_apps_handlers(app: web.Application, routes: ListOfRoutes):
    """
    Initialize sub application handlers

    :param app:
    :param routes: list of AbstractRoute subclasses
    """
    for route in routes:
        app.router.add_route(
            handler=route.handler,
            method=route.method,
            name=route.name,
            path=route.url_for().path
        )


class BaseProxy(abc.ABC):

    def __init__(self, dispatcher: Dispatcher, *,
                 loop: typing.Optional[AbstractEventLoop] = None):
        self.bot = dispatcher.bot
        self.dispatcher = dispatcher

        if isinstance(loop, AbstractEventLoop):
            if loop.is_closed():
                raise RuntimeError("The event loop, that you have passed on is closed")
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()

    @abc.abstractmethod
    def setup(self, **kwargs: typing.Any) -> typing.Any:
        """
        This method should establish the necessary for work

        """
        raise NotImplementedError


class TelegramWebhookProxy(BaseProxy):
    """
    Managing loading webhooks of aiogram together with QiwiWrapper

    """

    execution_path: str = "/bot/{token}"
    """You can override this parameter to change the default route path"""

    prefix: str = "/tg/webhooks"
    """ You can override the prefix for the application """

    def __init__(self, dispatcher: Dispatcher, sub_apps: typing.Optional[SubApps] = None):
        """

        :param dispatcher: instance of aiogram class Dispatcher
        :param sub_apps: list of tuples(prefix as string, web.Application)
        """
        from aiogram.dispatcher.webhook import WebhookRequestHandler  # type: ignore

        super(TelegramWebhookProxy, self).__init__(dispatcher)
        self._app: web.Application = web.Application()
        self._app.router.add_route(
            '*', self.execution_path, WebhookRequestHandler,
            name='webhook_handler'
        )
        self._app['BOT_DISPATCHER'] = self.dispatcher
        self.sub_apps: SubApps = sub_apps or []

    def setup(self, **kwargs) -> web.Application:
        """
        A method that connects the main application, and the proxy in the form of a telegram

        :param kwargs: keyword arguments, which contains application and host
        """
        main_app: web.Application = kwargs.pop("app")
        host: str = kwargs.pop("host")

        main_app.add_subapp(self.prefix, self._app)
        for prefix, sub_app, handlers in self.sub_apps:
            sub_app['bot'] = self.bot
            sub_app['dp'] = self.dispatcher
            _init_sub_apps_handlers(sub_app, handlers)
            main_app.add_subapp(prefix, sub_app)
        self._loop.run_until_complete(
            self.configure_webhook(host, **kwargs)
        )

        return self._app

    async def configure_webhook(self, host: str, **kwargs) -> typing.Any:
        """
        You can override this method to correctly setup webhooks with aiogram
        API method `set_webhook` like this: self.dispatcher.bot.set_webhook()

        """

        full_url = host + self.prefix

        if isinstance(self.execution_path, str):
            full_url += self.execution_path

        await self.dispatcher.bot.set_webhook(full_url, **kwargs)


class TelegramPollingProxy(BaseProxy):
    """
    Builtin telegram proxy.
    Allows you to use Telegram and QIWI webhooks together

    """

    def __init__(
            self,
            dispatcher: Dispatcher,
            loop: typing.Optional[AbstractEventLoop] = None,
            timeout: int = 20,
            relax: float = 0.1,
            limit=None,
            reset_webhook=None,
            fast: typing.Optional[bool] = True,
            error_sleep: int = 5,
            allowed_updates: typing.Optional[typing.List[str]] = None
    ):
        super(TelegramPollingProxy, self).__init__(dispatcher, loop=loop)
        self._allowed_updates = allowed_updates
        self._error_sleep = error_sleep
        self._fast = fast
        self._reset_webhook = reset_webhook
        self._timeout = timeout
        self._relax = relax
        self._limit = limit

    def setup(self, **kwargs: typing.Any):
        """
        Set up polling to run polling qiwi updates concurrently with aiogram

        :param kwargs: you can pass on loop as key/value parameter
        """
        loop = kwargs.pop("loop") or self._loop
        loop.create_task(self.dispatcher.start_polling(
            timeout=self._timeout,
            reset_webhook=self._reset_webhook,
            relax=self._relax,
            allowed_updates=self._allowed_updates,
            limit=self._limit,
            error_sleep=self._error_sleep,
            fast=self._fast
        ))



from datetime import timedelta, datetime

from aiohttp import ClientTimeout

DEFAULT_TIMEOUT = ClientTimeout(total=5 * 60)

USER_AGENT = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36' \
             '(KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36'

uncached = (
    'https://api.qiwi.com/partner/bill',
    '/sinap/api/v2/terms/'
)

DEFAULT_BILL_TIME = datetime.now() + timedelta(days=2)



from typing import List

from pydantic import Field

from glQiwiApi.types.base import Base
from glQiwiApi.types.basics import Sum


class Statistic(Base):
    """ object: Statistic """
    incoming: List[Sum] = Field(alias="incomingTotal")
    out: List[Sum] = Field(alias="outgoingTotal")


__all__ = [
    'Statistic'
]



from datetime import date
from typing import Optional

from pydantic import Field

from glQiwiApi.types.base import Base


class Identification(Base):
    """ object: Identification """
    identification_id: int = Field(..., alias="id")
    first_name: str = Field(..., alias="firstName")
    middle_name: str = Field(..., alias="middleName")
    last_name: str = Field(..., alias="lastName")
    birth_date: date = Field(..., alias="birthDate")
    passport: str
    inn: Optional[str]
    snils: Optional[str]
    oms: Optional[str]
    type: str



from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import Optional, Dict, Any

from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.utils.basics import check_api_method


class YooMoneyRouter(AbstractRouter):
    __head__ = 'https://yoomoney.ru'

    def setup_config(self) -> YooMoneySettings:
        return get_settings()

    @functools.lru_cache()
    def build_url(self, api_method: str, **kwargs: Any) -> str:
        check_api_method(api_method)
        tail_path: Optional[str] = getattr(self.config, api_method, None)
        pre_build_url = self.__head__ + tail_path
        return super()._format_url_kwargs(pre_build_url, **kwargs)


@functools.lru_cache()
def get_settings() -> YooMoneySettings:
    settings = YooMoneySettings()
    return settings


@dataclass
class YooMoneySettings:
    # Operations with token
    BUILD_URL_FOR_AUTH: str = '/oauth/authorize'
    GET_ACCESS_TOKEN: str = '/oauth/token'
    REVOKE_API_TOKEN: str = '/api/revoke'
    ACCOUNT_INFO: str = '/api/account-info'
    # Transactions
    TRANSACTIONS: str = '/api/operation-history'
    TRANSACTION_INFO: str = '/api/operation-details'
    # Payments
    PRE_PROCESS_PAYMENT: str = '/api/request-payment'
    PROCESS_PAYMENT: str = '/api/process-payment'
    ACCEPT_INCOMING_TRANSFER: str = '/api/incoming-transfer-accept'
    INCOMING_TRANSFER_REJECT: str = '/api/incoming-transfer-reject'

    def __init__(self):
        self.DEFAULT_YOOMONEY_HEADERS = {
            'Host': 'yoomoney.ru',
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        self.ERROR_CODE_NUMBERS = {
            400: "Ошибка, связанная с типом запроса к апи, "
                   "возможно вы передали недействительный API токен",
            401: "Указан несуществующий, просроченный, или отозванный токен",
            403: "Запрошена операция, на которую у токена нет прав",
            0: "Ошибка, связанная с использованием прокси или непредвиденные ошибки сервера"
        }
        self.content_and_auth = {
            'content_json': True, 'auth': True
        }

    DEFAULT_YOOMONEY_HEADERS: Optional[Dict[str, str]] = None

    ERROR_CODE_NUMBERS: Optional[Dict[int, str]] = None

    content_and_auth: Optional[Dict[str, bool]] = None


__all__ = ('YooMoneyRouter',)



from datetime import datetime
from typing import Union

from pydantic import Field, validator

from glQiwiApi.types.base import Base
from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class Interval(Base):
    """ object: Interval """
    date_from: datetime = Field(alias="dateFrom")
    date_till: datetime = Field(alias="dateTill")


class Limit(Base):
    """ object: Limit """
    currency: CurrencyModel
    rest: Union[float, int]
    max_limit: Union[float, int] = Field(alias="max")
    spent: Union[float, int]
    interval: Interval
    limit_type: str = Field(alias="type")

    @validator("currency", pre=True)
    def currency_validate(cls, v):
        if not isinstance(v, str):
            raise ValueError()
        return Currency.get(v)


__all__ = [
    'Limit'
]



import time
import typing

from glQiwiApi.core import BaseStorage
from glQiwiApi.core.constants import uncached
from glQiwiApi.types.basics import Cached, Attributes


class Storage(BaseStorage):
    """
    Deal with cache and data. Easy to use

    >>> storage = Storage(cache_time=5)
    >>> storage["hello_world"] = 5
    >>> print(storage["hello_world"])  # print 5

    """
    # Доступные критерии, по которым проходит валидацию кэш
    _validator_criteria = ('params', 'json', 'data', 'headers')

    __slots__ = ('data', '_cache_time')

    def __init__(self, *, cache_time: typing.Union[float, int]) -> None:
        """

        :param cache_time: Время кэширования в секундах
        """

        self.data: dict = {}
        self._cache_time = cache_time

    def clear(self, key: typing.Optional[str] = None, *,
              force: bool = False) -> typing.Any:
        """
        Method to delete element from the cache by key,
        or if force passed on its clear all data from the cache

        :param key: by this key to delete an element in storage
        :param force: If this argument is passed as True,
         the date in the storage will be completely cleared.

        """
        if force:
            return self.data.clear()
        return self.data.pop(key)

    def __setitem__(self, key, value) -> None:
        return self.update_data(value, key)

    def __getitem__(self, item) -> typing.Optional[typing.Any]:
        try:
            obj = self.data[item]
            if not self._is_expire(obj["cached_in"], item):
                return obj["data"]
            return None
        except KeyError:
            return None

    def _is_contain_uncached(self, value: typing.Optional[typing.Any]) -> bool:
        if self._cache_time < 0.1:
            return True

        for coincidence in uncached:
            try:
                if value.startswith(coincidence) or coincidence in value:  # type: ignore
                    return True
            except AttributeError:
                return False
        return False

    def convert_to_cache(
            self,
            result: typing.Any,
            kwargs: dict,
            status_code: typing.Union[str, int]
    ) -> Cached:
        """
        Method, which convert response of API to :class:`Cached`

        :param result: response data
        :param kwargs: key/value of request payload data
        :param status_code: status_code answer
        """
        value = kwargs.get("url")
        if not self._is_contain_uncached(value):
            return Cached(
                kwargs=Attributes.format(kwargs, self._validator_criteria),
                response_data=result,
                status_code=status_code,
                method=kwargs.get('method')
            )
        elif isinstance(value, str):
            if uncached[1] in value:
                self.clear(value, force=True)

    def update_data(self, obj_to_cache: typing.Any, key: typing.Any) -> None:
        """
        Метод, который добавляет результат запроса в кэш.

        >>> storage = Storage(cache_time=5)
        >>> storage.update_data(obj_to_cache="hello world", key="world")
        >>> storage["world"] = "hello_world"  # This approach is in priority and
        ...                                   # the same as on the line of code above

        :param obj_to_cache: объект для кэширования
        :param key: ключ, по которому будет зарезервирован этот кэш
        """
        if not self._is_contain_uncached(obj_to_cache):
            self.data[key] = {
                "data": obj_to_cache,
                "cached_in": time.monotonic(),
            }

    @staticmethod
    def _check_get_request(cached: Cached, kwargs: dict) -> bool:
        """ Method to check cached get requests"""
        if cached.method == 'GET':
            if kwargs.get('headers') == cached.kwargs.headers:
                if kwargs.get('params') == cached.kwargs.params:
                    return True
        return False

    def _is_expire(self, cached_in: float, key: typing.Any) -> bool:
        """ Method to check live cache time, and if it expired return True """
        if time.monotonic() - cached_in > self._cache_time:
            self.clear(key)
            return True
        return False

    def _validate_other(self, cached: Cached, kwargs: dict) -> bool:
        keys = (k for k in self._validator_criteria if k != 'headers')
        for key in keys:
            if getattr(cached.kwargs, key) == kwargs.get(key, ''):
                return True
        return False

    def validate(self, **kwargs) -> bool:
        """
        Метод, который по некоторым условиям
        проверяет актуальность кэша и в некоторых
        случая его чистит.

        :param kwargs:
        :return: boolean, прошел ли кэшированный запрос валидацию
        """
        # Если параметры и ссылка запроса совпадает
        cached = self[kwargs.get("url")]
        validation_kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if isinstance(cached, Cached):
            # Проверяем запрос методом GET на кэш
            if self._check_get_request(cached, validation_kwargs):
                return True
            elif self._validate_other(cached, validation_kwargs):
                return True

        return False



"""
Provides effortless work with YooMoney API using asynchronous requests.

"""
from datetime import datetime
from typing import List, Dict, Any, Union, Optional, Tuple

from pydantic import ValidationError

import glQiwiApi.utils.basics as api_helper
from glQiwiApi.core import (
    RequestManager,
    ToolsMixin,
    ContextInstanceMixin
)
from glQiwiApi.types import (
    AccountInfo,
    OperationType,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    IncomingTransaction
)
from glQiwiApi.types.basics import DEFAULT_CACHE_TIME
from glQiwiApi.utils.exceptions import NoUrlFound, InvalidData
from glQiwiApi.yoo_money.settings import YooMoneyRouter


class YooMoneyAPI(ToolsMixin, ContextInstanceMixin["YooMoneyAPI"]):
    """
    Класс, реализующий обработку запросов к YooMoney
    Удобен он тем, что не просто отдает json подобные объекты,
    а всё это конвертирует в pydantic модели.
    Для работы с данным классом, необходимо зарегистрировать токен,
    используя гайд на официальном github проекта

    """

    __slots__ = ("api_access_token", "_requests", "_router")

    def __init__(
            self,
            api_access_token: str,
            without_context: bool = False,
            cache_time: Union[float, int] = DEFAULT_CACHE_TIME,
            proxy: Optional[Any] = None
    ) -> None:
        """
        Конструктор принимает токен, полученный из класс метода
         get_access_token
        и специальный аттрибут without_context,

        :param api_access_token: апи токен для запросов
        :param without_context: bool, указывающая будет
         ли объект класса "глобальной" переменной
         или будет использована в async with контексте
        :param cache_time: Время кэширование запросов в секундах,
         по умолчанию 0, соответственно,
         запрос не будет использовать кэш по дефолту, максимальное время
         кэширование 60 секунд
        """
        self.api_access_token = api_access_token
        self._router = YooMoneyRouter()
        self._requests = RequestManager(
            without_context=without_context,
            messages=self._router.config.ERROR_CODE_NUMBERS,
            cache_time=cache_time,
            proxy=proxy
        )

        self.set_current(self)

    def _auth_token(self, headers: dict) -> Dict[Any, Any]:
        headers['Authorization'] = headers['Authorization'].format(
            token=self.api_access_token
        )
        return headers

    @property
    def request_manager(self) -> RequestManager:
        return self._requests

    @classmethod
    async def build_url_for_auth(
            cls,
            scope: List[str],
            client_id: str,
            redirect_uri: str = 'https://example.com'
    ) -> Optional[str]:
        """
        Метод для получения ссылки для
        дальнейшей авторизации и получения токена

        :param scope: OAuth2-авторизации приложения пользователем,
         права передаются списком.
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда попадет временный код, который нужен
         для получения основного токена
        :return: ссылку, по которой нужно перейти
         и сделать авторизацию через логин/пароль
        """
        router = YooMoneyRouter()
        headers = api_helper.parse_headers()
        params = {
            'client_id': client_id,
            'response_type': 'code',
            'redirect_uri': redirect_uri,
            'scope': " ".join(scope)
        }
        async for response in RequestManager(
                without_context=True,
                messages=router.config.ERROR_CODE_NUMBERS
        ).fast().fetch(
            url=router.build_url("BUILD_URL_FOR_AUTH"),
            headers=headers,
            data=params,
            method='POST'
        ):
            try:
                return api_helper.parse_auth_link(response.response_data)
            except IndexError:
                raise NoUrlFound(
                    'Не удалось найти ссылку для авторизации '
                    'в ответе от апи, проверьте client_id'
                )

    @classmethod
    async def get_access_token(
            cls,
            code: str,
            client_id: str,
            redirect_uri: str = 'https://example.com',
            client_secret: Optional[str] = None
    ) -> str:
        """
        Метод для получения токена для запросов к YooMoney API

        :param code: временный код, который был получен в методе base_authorize
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда попадет временный код,
         который нужен для получения основного токена
        :param client_secret: Секретное слово для проверки подлинности приложения.
         Указывается, если сервис зарегистрирован с проверкой подлинности.
        :return: YooMoney API TOKEN
        """
        router = YooMoneyRouter()
        headers = api_helper.parse_headers(content_json=True)
        params = {
            'code': code,
            'client_id': client_id,
            'grant_type': 'authorization_code',
            'redirect_uri': redirect_uri,
            'client_secret': client_secret
        }
        async for response in RequestManager(
                without_context=True,
                messages=router.config.ERROR_CODE_NUMBERS
        ).fast().fetch(
            url=router.build_url("GET_ACCESS_TOKEN"),
            headers=headers,
            data=params,
            method='POST',
            get_json=True
        ):
            return response.response_data.get('access_token')

    async def revoke_api_token(self) -> Optional[Dict[str, bool]]:
        """
        Метод для отмены прав токена, при этом все его права тоже пропадают
        Документация:
        https://yoomoney.ru/docs/wallet/using-api/authorization/revoke-access-token
        """
        headers = self._auth_token(
            api_helper.parse_headers(auth=True)
        )
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("REVOKE_API_TOKEN"),
                method='POST',
                headers=headers
        ):
            if response.status_code == 200:
                return {'success': True}

    @property
    async def account_info(self) -> AccountInfo:
        """
        Метод для получения информации об аккаунте пользователя
        Подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/account-info

        :return: объект AccountInfo
        """
        headers = self._auth_token(
            api_helper.parse_headers(
                **self._router.config.content_and_auth
            )
        )
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("ACCOUNT_INFO"),
                headers=headers,
                method='POST',
                get_json=True
        ):
            return AccountInfo.parse_obj(response.response_data)

    async def transactions(
            self,
            operation_types: Optional[
                Union[List[OperationType], Tuple[OperationType, ...]]
            ] = None,
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None,
            start_record: Optional[int] = None,
            records: int = 30,
            label: Optional[Union[str, int]] = None
    ) -> List[Operation]:
        """
        Подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/operation-history\n
        Метод позволяет просматривать историю операций (полностью или частично)
        в постраничном режиме.\n
        Записи истории выдаются в обратном хронологическом порядке:
        от последних к более ранним.\n
        Перечень типов операций, которые требуется отобразить.\n
        Возможные значения:
        DEPOSITION — пополнение счета (приход);\n
        PAYMENT — платежи со счета (расход);\n
        INCOMING(incoming-transfers-unaccepted) —
         непринятые входящие P2P-переводы любого типа.\n

        :param operation_types: Тип операций
        :param label: string.
         Отбор платежей по значению метки.
         Выбираются платежи, у которых указано заданное значение параметра
         label вызова request-payment.
        :param start_date: datetime	Вывести операции от момента времени
         (операции, равные start_date, или более поздние)
         Если параметр отсутствует, выводятся все операции.
        :param end_date: datetime	Вывести операции до момента времени
         (операции более ранние, чем end_date).
         Если параметр отсутствует, выводятся все операции.
        :param start_record: string
         Если параметр присутствует, то будут отображены операции, начиная
          с номера start_record.
         Операции нумеруются с 0. Подробнее про постраничный вывод списка
        :param records:	int	Количество запрашиваемых записей истории операций.
         Допустимые значения: от 1 до 100, по умолчанию — 30.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**self._router.config.content_and_auth)
        )

        if records <= 0 or records > 100:
            raise InvalidData(
                'Неверное количество записей. '
                'Кол-во записей, которые можно запросить,'
                ' находиться в диапазоне от 1 до 100 включительно'
            )

        data = {
            'records': records,
            'label': label
        }

        if operation_types:
            if all(
                    isinstance(
                        operation_type, OperationType
                    ) for operation_type in operation_types
            ):
                op_types = [
                    operation_type.value for operation_type in operation_types
                ]
                data.update({'type': ' '.join(
                    op_types)}
                )

        if isinstance(start_record, int):
            if start_record < 0:
                raise InvalidData('Укажите позитивное число')
            data.update({'start_record': start_record})

        if start_date:
            if not isinstance(start_date, datetime):
                raise InvalidData(
                    'Параметр start_date был передан неправильным типом данных'
                )
            data.update({'from': api_helper.datetime_to_str_in_iso(
                start_date, yoo_money_format=True
            )})

        if end_date:
            if not isinstance(end_date, datetime):
                raise InvalidData(
                    'Параметр end_date был передан неправильным типом данных'
                )
            data.update({'till': api_helper.datetime_to_str_in_iso(
                end_date, yoo_money_format=True
            )})

        async for response in self._requests.fast().fetch(
                url=self._router.build_url("TRANSACTIONS"),
                method='POST',
                headers=headers,
                data=self._requests.filter_dict(data),
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data.get('operations'),
                model=Operation
            )

    async def transaction_info(self, operation_id: str) -> OperationDetails:
        """
        Позволяет получить детальную информацию об операции из истории.
        Требуемые права токена: operation-details.
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/operation-details

        :param operation_id: Идентификатор операции
        """
        headers = self._auth_token(api_helper.parse_headers(
            **self._router.config.content_and_auth)
        )
        payload = {
            'operation_id': operation_id
        }
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("TRANSACTION_INFO"),
                headers=headers,
                data=payload,
                get_json=True
        ):
            return OperationDetails.parse_obj(response.response_data)

    async def _pre_process_payment(
            self,
            to_account: str,
            amount: Union[int, float],
            pattern_id: str = 'p2p',
            comment_for_history: Optional[str] = None,
            comment_for_receiver: Optional[str] = None,
            protect: bool = False,
            expire_period: int = 1
    ) -> PreProcessPaymentResponse:
        """
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/request-payment\n
        Создание платежа, проверка параметров и возможности приема
        платежа магазином или перевода средств на счет пользователя ЮMoney.\n
        Данный метод не рекомендуется использовать напрямую, гораздо
        проще использовать send.
        Требуемые права токена:
        to-account («идентификатор получателя», «тип идентификатора»)

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода
         (номер счета, номер телефона или email).
        :param amount: Сумма к получению
         (придет на счет получателя счет после оплаты).
        :param comment_for_history: Комментарий к переводу,
         отображается в истории отправителя.
        :param comment_for_receiver:	string	Комментарий к переводу,
         отображается получателю.
        :param protect: Значение параметра true — признак того,
         что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции
            и получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365.
            Необязательный параметр. По умолчанию 1.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**self._router.config.content_and_auth))
        payload = {
            'pattern_id': pattern_id,
            'to': to_account,
            'amount_due': amount,
            'comment': comment_for_history,
            'message': comment_for_receiver,
            'expire_period': expire_period,
            'codepro': protect
        }
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("PRE_PROCESS_PAYMENT"),
                headers=headers,
                data=payload,
                get_json=True
        ):
            try:
                return PreProcessPaymentResponse.parse_obj(
                    response.response_data
                )
            except ValidationError:
                msg = "Недостаточно денег для перевода или ошибка сервиса"
                raise self._requests.make_exception(
                    status_code=400,
                    message=msg
                )

    async def send(
            self,
            to_account: str,
            amount: Union[int, float],
            money_source: str = 'wallet',
            pattern_id: str = 'p2p',
            cvv2_code: str = '',
            card_type: Optional[str] = None,
            protect: bool = False,
            comment_for_history: Optional[str] = None,
            comment: Optional[str] = None,
            expire_period: int = 1
    ) -> Payment:
        """
        Метод для отправки денег на аккаунт или карту другого человека.
        Данная функция делает сразу 2 запроса, из-за этого
         вы можете почувствовать небольшую потерю производительности,
          вы можете использовать метод
        _pre_process_payment и получать объект PreProcessPaymentResponse,
        в котором есть информация о ещё неподтвержденном платеже\n
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/process-payment

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода
         (номер счета, номер телефона или email).
        :param amount: Сумма к получению
         (придет на счет получателя счет после оплаты). МИНИМАЛЬНАЯ СУММА 2.
        :param money_source: Запрашиваемый метод проведения платежа.
         wallet — со счета пользователя, если вы хотите использовать card,
         тогда нужно будет передать card_type
         для поиска карты в списке ваших банковских карт,
         а также опционально cvv2 код для проведения платежа
        :param comment_for_history: Комментарий к переводу,
         отображается в истории отправителя.
        :param card_type: Тип банковской карты, нужно заполнять,
         только если вы хотите списать средства с вашей карты
        :param cvv2_code: опционально, может быть не передан, однако,
         если для оплаты картой это требуется,
         параметр стоит передавать
        :param comment:	Комментарий к переводу, отображается получателю.
        :param protect: Значение параметра true — признак того,
         что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции и
            получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365.
            Необязательный параметр. По умолчанию 1.
        """
        if amount < 2:
            raise InvalidData(
                'Введите сумму, которая больше минимальной(2 и выше)'
            )
        url = self._router.build_url("PROCESS_PAYMENT")
        pre_payment = await self._pre_process_payment(
            to_account=to_account,
            amount=amount,
            comment_for_history=comment_for_history,
            comment_for_receiver=comment,
            expire_period=expire_period,
            protect=protect,
            pattern_id=pattern_id
        )

        headers = self._auth_token(
            api_helper.parse_headers(**self._router.config.content_and_auth))

        payload = {
            'request_id': pre_payment.request_id,
            'money_source': 'wallet'
        }
        expression = isinstance(pre_payment, PreProcessPaymentResponse)
        if money_source == 'card' and expression:
            if pre_payment.money_source.cards.allowed == 'true':
                if not isinstance(card_type, str):
                    cards = pre_payment.money_source.cards
                    payload.update({
                        'money_source': cards.items[0].item_id,
                        'csc': cvv2_code
                    })
                else:
                    cards = pre_payment.money_source.cards.items
                    for card in cards:
                        if card.item_type == card_type:
                            payload.update(
                                {
                                    'money_source': card.item_id,
                                    'csc': cvv2_code
                                },
                            )
        async for response in self._requests.fast().fetch(
                url=url,
                method='POST',
                headers=headers,
                data=payload,
                get_json=True
        ):
            return Payment.parse_obj(
                response.response_data
            ).initialize(pre_payment.protection_code)

    @property
    async def balance(self) -> float:
        """Метод для получения баланса на кошельке yoomoney"""
        return (await self.account_info).balance

    async def accept_incoming_transaction(
            self,
            operation_id: str,
            protection_code: str
    ) -> IncomingTransaction:
        """
        Прием входящих переводов, защищенных кодом протекции,
        если вы передали в метод send параметр protect
        Количество попыток приема
        входящего перевода с кодом протекции ограничено.
        При исчерпании количества попыток, перевод автоматически отвергается
        (перевод возвращается отправителю).
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-accept

        :param operation_id: Идентификатор операции,
         значение параметра operation_id ответа метода history()
        :param protection_code: Код протекции. Строка из 4-х десятичных цифр.
         Указывается для входящего перевода, защищенного кодом протекции.
         Для переводов до востребования отсутствует.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**self._router.config.content_and_auth))
        payload = {
            'operation_id': operation_id,
            'protection_code': protection_code
        }
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("ACCEPT_INCOMING_TRANSFER"),
                headers=headers,
                data=payload,
                method='POST',
                get_json=True
        ):
            return IncomingTransaction.parse_obj(
                response.response_data
            )

    async def reject_transaction(self, operation_id: str) -> Dict[str, str]:
        """
        Отмена входящих переводов, защищенных кодом протекции, если вы передали
         в метод send параметр protect,
        и переводов до востребования. \n
        При отмене перевода он возвращается отправителю. \n
        Требуемые права токена: incoming-transfers

        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-reject

        :param operation_id: Идентификатор операции, значение параметра
         operation_id ответа метода history().
        :return: словарь в json формате с ответом от апи
        """
        headers = self._auth_token(
            api_helper.parse_headers(**self._router.config.content_and_auth))
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("INCOMING_TRANSFER_REJECT"),
                headers=headers,
                data={'operation_id': operation_id},
                method='POST',
                get_json=True
        ):
            return response.response_data

    async def check_transaction(
            self,
            amount: Union[int, float],
            transaction_type: str = 'in',
            comment: Optional[str] = None,
            rows_num: int = 100,
            recipient: Optional[str] = None
    ) -> bool:
        """
        Метод для проверки транзакции.\n
        Данный метод использует self.transactions(rows_num=rows_num)
         для получения платежей.\n
        Для небольшой оптимизации вы можете уменьшить rows_num задав его,
        однако это не гарантирует правильный результат

        :param amount: сумма платежа
        :param transaction_type: тип платежа
        :param recipient: номер получателя
        :param rows_num: кол-во платежей, которое будет проверяться
        :param comment: комментарий, по которому будет проверяться транзакция
        :return: bool, есть ли такая транзакция в истории платежей
        """
        # Generate types of transactions for request
        types = [
            OperationType.DEPOSITION if transaction_type == 'in' else OperationType.PAYMENT]
        # Get transactions by params
        transactions = await self.transactions(
            operation_types=types,
            records=rows_num
        )
        for txn in transactions:
            # Get details of transaction to check it later
            detail = await self.transaction_info(txn.operation_id)

            # Parse amount and comment,
            # because the parameters depend on the type of transaction
            amount_, comment_ = api_helper.parse_amount(
                transaction_type,
                detail
            )
            checked = api_helper.check_params(
                amount=amount,
                transaction_type=transaction_type,
                amount_=amount_,
                txn=detail
            )
            if checked:
                if txn.status == 'success':
                    if comment_ == comment and detail.sender == recipient:
                        return True
                    elif isinstance(comment, str) and isinstance(recipient, str):
                        continue
                    elif comment_ == comment:
                        return True

        return False



from typing import Dict, Optional, Any, Union, List

import aiohttp
from aiohttp.typedefs import LooseCookies

from glQiwiApi.core.basic_requests_api import HttpXParser, _ProxyType
from glQiwiApi.core.storage import Storage
from glQiwiApi.types import Response
from glQiwiApi.types.basics import Cached, DEFAULT_CACHE_TIME


class RequestManager(HttpXParser):
    """
    Deal with :class:`Storage`,
    caching queries and managing stable work of sending requests

    """
    __slots__ = (
        'without_context', 'messages', '_cache', '_should_reset_connector',
        '_connector_type', '_connector_init', '_proxy'
    )

    def __init__(
            self,
            without_context: bool = False,
            messages: Optional[Dict[str, str]] = None,
            cache_time: Union[float, int] = DEFAULT_CACHE_TIME,
            proxy: Optional[_ProxyType] = None
    ) -> None:
        super(RequestManager, self).__init__(proxy=proxy, messages=messages)

        self.without_context: bool = without_context
        self._cache: Storage = Storage(cache_time=cache_time)

    def reset_cache(self) -> None:
        """ Clear all cache in storage """
        self._cache.clear(force=True)

    async def make_request(self, **kwargs) -> Union[Response, Cached]:
        """ The user-friendly method that allows sending requests to any URL  """
        return await super(RequestManager, self)._make_request(**kwargs)

    async def _make_request(self,
                            url: str,
                            get_json: bool = False,
                            method: str = 'POST',
                            set_timeout: bool = True,
                            cookies: Optional[LooseCookies] = None,
                            json: Optional[dict] = None,
                            data: Optional[Dict[str, Union[
                                str, int, List[
                                    Union[str, int]
                                ]]]
                            ] = None,
                            headers: Optional[dict] = None,
                            params: Optional[
                                Dict[str, Union[str, int, List[
                                    Union[str, int]
                                ]]]
                            ] = None,
                            get_bytes: bool = False,
                            **kwargs) -> Union[Response, Cached]:
        """ Send request to service(API) """
        request_args = {k: v for k, v in locals().items() if not isinstance(v, type(self))}
        # Получаем текущий кэш используя ссылку как ключ
        response: Optional[Any] = self._cache[url]
        if not self._cache.validate(**request_args):
            response = await super()._make_request(
                url=url,
                method=method,
                get_json=get_json,
                get_bytes=get_bytes,
                cookies=cookies,
                headers=headers,
                params=params,
                set_timeout=set_timeout,
                data=data,
                json=json
            )
        # Проверяем, не был ли запрос в кэше, если нет,
        # то проверяем статус код и если он не 200 - выбрасываем ошибку
        if not isinstance(response, Cached) and isinstance(response, Response):
            if self.without_context:
                await self._close_session()
            if response.status_code != 200:
                raise self.make_exception(
                    response.status_code,
                    traceback_info=response.response_data
                )
            else:
                self._cache_all(response, **request_args)

        return response

    async def _close_session(self):
        await super(RequestManager, self).close()

    async def close(self) -> None:
        """ Close aiohttp session and reset cache data """
        await super(RequestManager, self).close()
        self.reset_cache()

    def _cache_all(self, response: Response, **kwargs: Any):
        resolved: Cached = self._cache.convert_to_cache(
            result=response.response_data,
            kwargs=kwargs,
            status_code=response.status_code
        )
        self._cache[kwargs["url"]] = resolved

    @property
    def is_session_closed(self) -> bool:
        if isinstance(self._session, aiohttp.ClientSession):
            if not self._session.closed:
                return True
        return False

    @classmethod
    def filter_dict(cls, dictionary: dict) -> dict:
        """
        Pop NoneType values and convert everything to str, designed?for=params

        :param dictionary: source dict
        :return: filtered dict
        """
        return {k: str(v) for k, v in dictionary.items() if v is not None}



import json
from typing import Optional, Union, Dict, Any

from aiohttp import RequestInfo
from pydantic import BaseModel


class NoUrlFound(Exception):
    """Данная ошибка возникает при неправильной авторизации yoomoney"""


class RequestProxyError(Exception):
    """Возникает, если были переданы неправильные параметры запроса"""


ProxyError = Exception()


class InvalidCardNumber(Exception):
    """
    Ошибка, при передаче номера карты в неправильном формате

    """


class RequestAuthError(Exception):
    """
    Ошибка при неправильной аутентификации POST or GET data

    """


class InvalidToken(Exception):
    """
    Ошибка, возникающая, если был передан неверный токен

    """


class InvalidData(TypeError):
    """
    Ошибка возникает если были переданы или получены невалидные данные

    """


class NoUpdatesToExecute(Exception):
    """
    Ошибка возбуждается если при полинге нет транзакций, чтобы обрабатывать

    """


class RequestInfoModel(BaseModel):
    method: str
    url: str
    real_url: str


class ExceptionTraceback(BaseModel):
    status_code: int
    msg: Optional[str] = None
    additional_info: Optional[str] = None
    request_info: Optional[RequestInfoModel] = None


class RequestError(Exception):
    """
    Возникает при ошибках сервиса или неправильной передаче параметров

    """

    def __init__(
            self,
            message: Optional[str],
            status_code: Union[str, int],
            additional_info: Optional[str] = None,
            traceback_info: Optional[Union[RequestInfo, str, bytes, dict]] = None
    ) -> None:
        super(RequestError, self).__init__()
        self.message = message
        self.status_code = status_code
        self.additional_info = additional_info
        self.traceback_info = traceback_info

    def __str__(self) -> str:
        resp = "code={sc} doc={msg}, additional_info={info}"""
        return resp.format(
            sc=self.status_code,
            msg=self.message,
            info=self.additional_info
        )

    def __repr__(self) -> str:
        return self.__str__()

    def to_model(self) -> ExceptionTraceback:
        """ Convert exception to :class:`ExceptionTraceback` """
        if not isinstance(self.traceback_info, RequestInfo):
            raise TypeError("Cannot convert exception to `ExceptionTraceback`, because "
                            "this method require `RequestInfo` object")
        return ExceptionTraceback(
            status_code=self.status_code,
            msg=self.message,
            additional_info=self.additional_info,
            request_info=RequestInfoModel(
                method=self.traceback_info.method,
                url=self.traceback_info.url.__str__(),
                real_url=self.traceback_info.real_url.__str__()
            )
        )

    def _make_json_without_request_info(self) -> Dict[str, Any]:
        return {
            "code": self.status_code,
            "msg": self.message,
            "additional_info": self.additional_info,
            "traceback_info": self.traceback_info
        }

    def json(self, indent: int = 4, **dump_kw) -> str:
        """
        Method, that makes json format from traceback

        :param indent:
        :param dump_kw:
        """
        if isinstance(self.traceback_info, RequestInfo):
            info = self.to_model().dict(exclude_none=True)
        else:
            info = self._make_json_without_request_info()
        return json.dumps(info, indent=indent, ensure_ascii=False, **dump_kw)


__all__ = (
    'InvalidData',
    'NoUrlFound',
    'RequestAuthError',
    'RequestProxyError',
    'ProxyError',
    'InvalidCardNumber',
    'InvalidToken',
    'RequestError',
    'NoUpdatesToExecute'
)



"""Main model: Terminal"""
from typing import Optional

from pydantic import Field

from glQiwiApi.types.base import Base


class Coordinate(Base):
    """Object: coordinate"""

    latitude: float = Field(..., alias="latitude")
    longitude: float = Field(..., alias="longitude")
    precision: int = Field(..., alias="precision")


class Terminal(Base):
    """Object: Terminal"""

    terminal_id: int = Field(..., alias="terminalId")
    ttp_id: int = Field(..., alias="ttpId")
    last_active: str = Field(..., alias="lastActive")
    count: int = Field(..., alias="count")
    address: str = Field(..., alias="address")
    verified: bool = Field(..., alias="verified")
    label: str = Field(..., alias="label")
    description: Optional[str] = Field(type(None), alias="description")
    cash_allowed: bool = Field(..., alias="cashAllowed")
    card_allowed: bool = Field(..., alias="cardAllowed")
    identification_type: int = Field(..., alias="identificationType")
    coordinate: Coordinate = Field(..., alias="coordinate")


__all__ = ("Terminal",)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glQiwiApi',
 'glQiwiApi.core',
 'glQiwiApi.core.builtin',
 'glQiwiApi.core.web_hooks',
 'glQiwiApi.qiwi',
 'glQiwiApi.types',
 'glQiwiApi.types.qiwi_types',
 'glQiwiApi.types.yoomoney_types',
 'glQiwiApi.utils',
 'glQiwiApi.yoo_money']

package_data = \
{'': ['*'],
 'glQiwiApi.core': ['.mypy_cache/*',
                    '.mypy_cache/3.8/*',
                    '.mypy_cache/3.8/_typeshed/*',
                    '.mypy_cache/3.8/aiohttp/*',
                    '.mypy_cache/3.8/async_timeout/*',
                    '.mypy_cache/3.8/asyncio/*',
                    '.mypy_cache/3.8/attr/*',
                    '.mypy_cache/3.8/builtin/*',
                    '.mypy_cache/3.8/chardet/*',
                    '.mypy_cache/3.8/collections/*',
                    '.mypy_cache/3.8/concurrent/*',
                    '.mypy_cache/3.8/concurrent/futures/*',
                    '.mypy_cache/3.8/ctypes/*',
                    '.mypy_cache/3.8/email/*',
                    '.mypy_cache/3.8/html/*',
                    '.mypy_cache/3.8/http/*',
                    '.mypy_cache/3.8/importlib/*',
                    '.mypy_cache/3.8/json/*',
                    '.mypy_cache/3.8/logging/*',
                    '.mypy_cache/3.8/loguru/*',
                    '.mypy_cache/3.8/multidict/*',
                    '.mypy_cache/3.8/multiprocessing/*',
                    '.mypy_cache/3.8/os/*',
                    '.mypy_cache/3.8/urllib/*',
                    '.mypy_cache/3.8/web_hooks/*',
                    '.mypy_cache/3.8/yarl/*'],
 'glQiwiApi.core.builtin': ['.mypy_cache/*',
                            '.mypy_cache/3.8/*',
                            '.mypy_cache/3.8/_typeshed/*',
                            '.mypy_cache/3.8/aiohttp/*',
                            '.mypy_cache/3.8/async_timeout/*',
                            '.mypy_cache/3.8/asyncio/*',
                            '.mypy_cache/3.8/attr/*',
                            '.mypy_cache/3.8/chardet/*',
                            '.mypy_cache/3.8/collections/*',
                            '.mypy_cache/3.8/concurrent/*',
                            '.mypy_cache/3.8/concurrent/futures/*',
                            '.mypy_cache/3.8/ctypes/*',
                            '.mypy_cache/3.8/email/*',
                            '.mypy_cache/3.8/glQiwiApi/core/*',
                            '.mypy_cache/3.8/glQiwiApi/core/builtin/*',
                            '.mypy_cache/3.8/glQiwiApi/core/web_hooks/*',
                            '.mypy_cache/3.8/glQiwiApi/types/*',
                            '.mypy_cache/3.8/glQiwiApi/types/qiwi_types/*',
                            '.mypy_cache/3.8/glQiwiApi/types/yoomoney_types/*',
                            '.mypy_cache/3.8/glQiwiApi/utils/*',
                            '.mypy_cache/3.8/html/*',
                            '.mypy_cache/3.8/http/*',
                            '.mypy_cache/3.8/importlib/*',
                            '.mypy_cache/3.8/json/*',
                            '.mypy_cache/3.8/logging/*',
                            '.mypy_cache/3.8/loguru/*',
                            '.mypy_cache/3.8/multidict/*',
                            '.mypy_cache/3.8/multiprocessing/*',
                            '.mypy_cache/3.8/os/*',
                            '.mypy_cache/3.8/pydantic/*',
                            '.mypy_cache/3.8/unittest/*',
                            '.mypy_cache/3.8/urllib/*',
                            '.mypy_cache/3.8/yarl/*'],
 'glQiwiApi.core.web_hooks': ['.mypy_cache/*',
                              '.mypy_cache/3.8/*',
                              '.mypy_cache/3.8/_typeshed/*',
                              '.mypy_cache/3.8/collections/*',
                              '.mypy_cache/3.8/importlib/*',
                              '.mypy_cache/3.8/os/*']}

install_requires = \
['aiofiles>=0.6.0,<0.7.0',
 'aiohttp>=3.7.4post0,<4.0.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic==1.8.2',
 'pytz==2021.1',
 'wheel>=0.36.2,<0.37.0']

extras_require = \
{':python_version < "3.7"': ['contextvars>=2.4,<3.0']}

setup_kwargs = {
    'name': 'glqiwiapi',
    'version': '1.0.2',
    'description': "The fast and lightweight wrapper of QIWI and YooMoney API's",
    'long_description': '<h2 align="center">\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/master/docs/static/logo.png" width="200"></img>\n\n\n[![PyPI version](https://img.shields.io/pypi/v/glQiwiApi.svg)](https://pypi.org/project/glQiwiApi/) [![Python](https://img.shields.io/badge/Python-3.7+-blue)](https://www.python.org/downloads/) [![Code Quality Score](https://www.code-inspector.com/project/20780/score/svg)](https://frontend.code-inspector.com/public/project/20780/glQiwiApi/dashboard) ![Code Grade](https://www.code-inspector.com/project/20780/status/svg) ![Downloads](https://img.shields.io/pypi/dm/glQiwiApi) ![docs](https://readthedocs.org/projects/pip/badge/?version=latest)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/GLEF1X/glQiwiApi.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/GLEF1X/glQiwiApi/context:python) [![CodeFactor](https://www.codefactor.io/repository/github/glef1x/glqiwiapi/badge)](https://www.codefactor.io/repository/github/glef1x/glqiwiapi)\n\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/master/demo.gif"/>\n</h2>\n\n## 🌎Official api resources:\n\n* 🎓 __Docs: [here](https://glqiwiapi.readthedocs.io/en/master/index.html)__\n* 🖱️ __Developer\n  contacts: [![Dev-Telegram](https://img.shields.io/badge/Telegram-blue.svg?style=flat-square&logo=telegram)](https://t.me/GLEF1X)__\n\n### 📣Why glQiwiApi?\n\n* :boom:__It\'s working faster than other async libraries for qiwi__\n* :dizzy:__Frequent updates and edits__\n* :innocent: __The library developer will always help you with any problems you\n  might encounter__\n\n### 💾Installation\n\n```bash\npip install glQiwiApi\n```\n\n---\n\n## 🐦Dependencies\n\n| Library | Description                                    |\n|:-------:|:----------------------------------------------:|\n|aiohttp  | default http server                            |\n|aiofiles | saving receipts in pdf                         |\n|uvloop   | Optional(can boost API), but work only on Linux|\n|pydantic | Json data validator. Very fast instead of custom|\n|loguru   | library which aims to bring enjoyable logging in Python|\n\n---\n\n## 🧸Dive-in Examples\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n\nasync def main():\n    # If you want to use qiwi wrapper without async context just \n    # pass on "without_context=True"\n    wallet = QiwiWrapper(\n        api_access_token=\'your_token\',\n        phone_number=\'+number\',\n        without_context=True\n    )\n    print((await wallet.get_balance()).amount)\n    # OR(x3 performance boost with async context,\n    # because it use only 1 aiohttp session to get response for all requests\n    # in async with context manager)\n    async with QiwiWrapper(api_access_token=\'your_token\') as w:\n        w.phone_number = \'+number\'\n        # Данным вызовом вы получите текущий баланс кошелька. \n        print((await w.get_balance()).amount)\n\n\n# Also you can use it like here\nmy_wallet = QiwiWrapper(\n    api_access_token=\'your_token\',\n    phone_number=\'+phone_number\'\n)\n\n\nasync def main_2():\n    async with my_wallet as w:\n        print(await w.get_balance())\n\n\nasyncio.run(main())\n```\n\n---\n\n## 🌀Checking transactions\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n\nasync def main():\n    async with QiwiWrapper(api_access_token=\'your_token\') as w:\n        w.phone_number = \'+number\'\n        # Таким образом мы проверим, была ли транзакция на сумму 999 рублей с комментарием\n        # \'I like glQiwiApi!\' и отправителем с номером +7904832168\n        is_paid = await w.check_transaction(\n            amount=999,\n            comment=\'I like glQiwiApi!\',\n            sender_number=\'+7904832168\'\n        )\n        print(is_paid)\n\n\nasyncio.run(main())\n```\n\n## 🌱Create & check p2p bills\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n\nasync def main():\n  # You can pass on only p2p tokens, if you want to use only p2p api\n  async with QiwiWrapper(\n          secret_p2p="your_secret_p2p"\n  ) as w:\n    # Таким образом можно создать p2p счет\n    # В примере указан счёт на 1 рубль с комментарием some_comment\n    bill = await w.create_p2p_bill(\n      amount=1,\n      comment=\'my_comm\'\n    )\n    # Проверка на статус "оплачено" созданного p2p счёта\n    if (await w.check_p2p_bill_status(bill_id=bill.bill_id)) == \'PAID\':\n      print(\'Успешно оплачено\')\n    else:\n      print(\'Транзакция не найдена\')\n    # Или, начиная с версии апи 0.2.0\n    print(await bill.paid)  # This will print you bool answer\n\n\nasyncio.run(main())\n```\n\n#### That will create a form like that\n\n![form](https://i.ibb.co/T0C5RYz/2021-03-21-14-58-33.png)\n\n## ⛏Send to another wallet & get receipt(получение чека транзакции)\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n\nasync def main():\n    async with QiwiWrapper(api_access_token="token") as w:\n        w.phone_number = "+number"\n        # Так выглядит перевод на другой киви кошелек\n        # в примере перевод будет на номер +7904832168 с комментарием "На шоколадку" и суммой 1 рубль\n        trans_id = await w.to_wallet(\n            to_number=\'+7904832168\',\n            comment=\'На шоколадку\',\n            trans_sum=1\n        )\n        # В данном примере мы сохраним чек в директории, где вы запускаете скрипт как my_receipt.pdf\n        await w.get_receipt(\n            transaction_id=trans_id,\n            transaction_type=\'OUT\',\n            file_path=\'my_receipt\'\n        )\n\n\nasyncio.run(main())\n\n```\n\n## 🌟Webhooks & handlers\n\n```python\n\nfrom glQiwiApi import QiwiWrapper, types\nfrom glQiwiApi.utils import executor\n\nwallet = QiwiWrapper(\n    api_access_token=\'token from https://qiwi.com/api/\',\n    secret_p2p=\'secret token from https://qiwi.com/p2p-admin/\'\n)\n\n\n@wallet.transaction_handler(lambda event: ...)\nasync def get_transaction(event: types.WebHook):\n    print(event)\n\n\n@wallet.bill_handler()\nasync def fetch_bill(notification: types.Notification):\n    print(notification)\n\n    \nexecutor.start_webhook(wallet, port=80)\n```\n\n## 🧑🏻🔬Polling updates\n```python\nimport datetime\n\nfrom glQiwiApi import QiwiWrapper, types\nfrom glQiwiApi.utils import executor\n\napi_access_token = "your token"\nphone_number = "your number"\n\nwallet = QiwiWrapper(\n    api_access_token=api_access_token,\n    phone_number=phone_number\n)\n\n\n@wallet.transaction_handler()\nasync def my_first_handler(update: types.Transaction):\n    assert isinstance(update, types.Transaction)\n\n\ndef on_startup(wrapper: QiwiWrapper):\n    wrapper.dispatcher.logger.info("This message logged on startup")\n\n\nexecutor.start_polling(\n    wallet,\n    on_startup=on_startup\n)\n```\n\n## 💳Send to card & check commission\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n\nasync def main():\n    async with QiwiWrapper(api_access_token="token") as w:\n        w.phone_number = "+number"\n        # Так можно отправлять средства на карты разных банков, получая при этом айди транзакции\n        trans_id = await w.to_card(\n            trans_sum=1,\n            to_card=\'4731219185432123\'\n        )\n        print(trans_id)\n        # Так можно предварительно расчитать комиссию за транзакцию\n        commission = await w.commission(\n            to_account=\'4731219185432123\',\n            pay_sum=1\n        )\n        print(commission.qiwi_commission.amount)\n\n\nasyncio.run(main())\n\n```\n\n## 🚀Query caching (beta)\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper\n\n# Кэширование по умолчанию отключено, так как\n# эта функция все ещё находиться в бета тестировании и\n# константа DEFAULT_CACHE_TIME = 0, чтобы это исправить и включить кэширование\n# нужно передать cache_time в конструктор класса QiwiWrapper\n# или YooMoneyAPI\nwallet = QiwiWrapper(\n    # Токен, полученный с https://qiwi.com/api\n    api_access_token=\'token\',\n    # Номер вашего мобильного номера с "+"\n    phone_number=\'+phone_number\',\n    # Время кэширование запроса в секундах(пока ещё в бета тестировании)\n    cache_time=5\n)\n\n\nasync def cache_test():\n    async with wallet:\n        # Результат заноситься в кэш\n        print(await wallet.transactions(rows_num=50))\n        # Этот запрос возьмется из кэша\n        print(await wallet.transactions(rows_num=50))\n\n        # Запросы ниже не будут браться из кэша,\n        # причиной тому есть разница в параметрах запроса\n        # Результат все также заноситься в кэш\n        print(len(await wallet.transactions(rows_num=30)) == 30)  # True\n        # Однако, повторный запрос к апи будет выполнен, поскольку\n        # при попытке взятие результата из кэша валидатор сравнивает\n        # параметры запроса, если они не совпадают, то\n        # кэш игнорируется\n        # Повторный запрос к апи\n        print(len(await wallet.transactions(rows_num=10)) == 10)  # True\n\n\nasyncio.run(cache_test())\n\n\n```\n\n## ⚠️Handling exceptions\n\n```python\nimport asyncio\n\nfrom glQiwiApi import QiwiWrapper, RequestError\n\n\nasync def main():\n    async with QiwiWrapper(api_access_token=\'your_token\') as w:\n        w.phone_number = \'+number\'\n        try:\n            await w.to_card(to_card="some_card", trans_sum=2)\n        except RequestError as ex:\n            # Its give u full traceback from api if response was bad\n            print(ex.json())\n\n\nasyncio.run(main())\n```\n\n---\n\n## 🗺QIWI terminals\n\n__glQiwiApi covers qiwi\'s MAPS api in QiwiMaps class__\n\n---\n\n# YooMoney API\n\n## Important. How to get YooMoney access token\n\n+ #### Регистрируем своё приложение в YooMoney по ссылке: [click](https://yoomoney.ru/myservices/new)\n\n![yoo_money_register_app](https://i.imgur.com/Mu6R8Po.png)\n\n+ #### Получаем client_id после регистрации и далее используем уже YooMoneyAPI:\n\n```python\nimport asyncio\n\nfrom glQiwiApi import YooMoneyAPI\n\n\nasync def get_url_to_auth() -> None:\n    # Получаем ссылку для авторизации, переходим по ней, если получаем invalid_request или какую-то ошибку\n    # значит либо неправильно передан scope параметр, нужно уменьшить список прав или попробовать пересоздать приложение\n    print(await YooMoneyAPI.build_url_for_auth(\n        # Для платежей, проверки аккаунта и истории платежей, нужно указать scope=["account-info", "operation-history", "operation-details", "payment-p2p"]\n        scope=["account-info", "operation-history"],\n        client_id=\'айди, полученный при регистрации приложения выше\',\n        redirect_uri=\'ссылка, указаная при регистрации выше в поле Redirect URI\'\n    ))\n\n\nasyncio.run(get_url_to_auth())\n```\n\n+ #### Теперь нужно получить временный код и МАКСИМАЛЬНО БЫСТРО получить токен, используя class method YooMoneyAPI:\n\n![reg](https://i2.paste.pics/7660ed1444d1b3fc74b08128c74dbcd4.png?trs=9bfa3b1c0203c2ffe9982e7813a27700d047bfbc7ed23b79b99c1c4ffdd34995)\n\n```python\nimport asyncio\n\nfrom glQiwiApi import YooMoneyAPI\n\n\nasync def get_token() -> None:\n    print(await YooMoneyAPI.get_access_token(\n        code=\'код полученный из ссылки, как на скрине выше\',\n        client_id=\'айди приложения, полученое при регистрации\',\n        redirect_uri=\'ссылка, указанная при регистрации\'\n    ))\n\n\nasyncio.run(get_token())\n```\n\n---\n\n## Dive-in Examples\n\n```python\nimport asyncio\n\nfrom glQiwiApi import YooMoneyAPI\n\nTOKEN = \'some_token\'\n\n\nasync def main():\n    wallet = YooMoneyAPI(\n        api_access_token=TOKEN,\n        without_context=True\n    )\n    transactions = await wallet.transactions()\n    print(transactions)\n    # OR(x3 performance boost)\n    async with YooMoneyAPI(api_access_token=TOKEN) as w:\n        print(await w.transactions(records=50))\n\n\nasyncio.run(main())\n```\n\n## Send money to another wallet and checking this transaction\n\n```python\n\nimport asyncio\n\nfrom glQiwiApi import YooMoneyAPI\n\nTOKEN = \'your_token\'\n\n\nasync def main():\n    w = YooMoneyAPI(TOKEN)\n    async with w:\n        # Так вы можете отослать средства на другой счет, в примере это перевод на аккаунт 4100116602400968\n        # на сумму 2 рубля с комментарием "I LOVE glQiwiApi"\n        payment = await w.send(\n            to_account=\'4100116602400968\',\n            comment=\'I LOVE glQiwiApi\',\n            amount=2\n        )\n        # Опционально, так вы можете проверить транзакцию, поступила ли она человеку на счёт\n        print(await w.check_transaction(amount=2, comment=\'I LOVE glQiwiApi\',\n                                        transaction_type=\'out\'))\n\n\nasyncio.run(main())\n\n```\n\n## Fetch account info\n\n```python\nimport asyncio\n\nfrom glQiwiApi import YooMoneyAPI\n\nTOKEN = \'your_token\'\n\n\nasync def main():\n    w = YooMoneyAPI(TOKEN)\n    async with w:\n        # Так вы получаете информацию об аккаунте в виде объекта AccountInfo\n        account_info = await w.account_info\n        print(account_info.account_status)\n        print(account_info.balance)\n\n\nasyncio.run(main())\n\n```\n',
    'author': 'Gleb Garanin',
    'author_email': 'glebgar567@gmail.com',
    'maintainer': 'GLEF1X',
    'maintainer_email': 'glebgar567@gmail.com',
    'url': 'https://github.com/GLEF1X/glQiwiApi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from typing import Optional, Union

from pydantic import Field

from glQiwiApi.types.base import Base
from glQiwiApi.types.basics import Sum


class Provider(Base):
    """ object: Provider """
    id: Optional[int] = None
    """ID провайдера в QIWI Wallet"""

    short_name: Optional[str] = Field(default=None, alias="shortName")
    """краткое наименование провайдера"""

    long_name: Optional[str] = Field(alias="longName", default=None)
    """развернутое наименование провайдера"""

    logo_url: Optional[str] = Field(alias="logoUrl", default=None)
    """ссылка на логотип провайдера"""

    description: Optional[str] = None
    """описание провайдера (HTML)"""

    keys: Optional[Union[str, list]] = None
    """список ключевых слов"""

    site_url: Optional[str] = Field(alias="siteUrl", default=None)
    """сайт провайдера"""


class Transaction(Base):
    """ object: Transaction """
    transaction_id: int = Field(alias="txnId")
    """	ID транзакции в сервисе QIWI Кошелек"""

    person_id: int = Field(alias="personId")
    """Номер кошелька"""

    date: datetime
    """
    Для запросов истории платежей - Дата/время платежа,
    во временной зоне запроса (см. параметр startDate).
    Для запросов данных о транзакции - Дата/время платежа, время московское
    """

    error_code: Optional[int] = Field(alias="errorCode", default=None)
    """Код ошибки платежа"""

    error: Optional[str] = None
    """Описание ошибки"""

    status: str
    """
    Статус платежа. Возможные значения:
    WAITING - платеж проводится,
    SUCCESS - успешный платеж,
    ERROR - ошибка платежа.
    """

    type: str
    """
    Тип платежа. Возможные значения:
    IN - пополнение,
    OUT - платеж,
    QIWI_CARD - платеж с карт QIWI (QVC, QVP).
    """

    status_text: str = Field(alias="statusText")
    """Текстовое описание статуса платежа"""

    trm_transaction_id: str = Field(alias="trmTxnId")
    """Клиентский ID транзакции"""

    to_account: str = Field(alias="account")
    """
    Для платежей - номер счета получателя.
    Для пополнений - номер отправителя,
    терминала или название агента пополнения кошелька
    """

    sum: Sum
    """Данные о сумме платежа или пополнения."""

    commission: Sum
    """Данные о комиссии"""

    total: Sum
    """Общие данные о платеже в формате объекта Sum"""

    provider: Provider
    """Данные о провайдере."""

    source: Optional[Provider] = None
    """Служебная информация"""

    comment: Optional[str] = None
    """	Комментарий к платежу"""

    currency_rate: int = Field(alias="currencyRate")
    """Курс конвертации (если применяется в транзакции)"""

    _extras: Optional[dict] = Field(..., alias="extras")
    """Служебная информация"""

    _cheque_ready: bool = Field(..., alias="chequeReady")
    """Специальное поле"""

    _bank_document_available: bool = Field(..., alias="bankDocumentAvailable")
    """Специальное поле"""

    _repeat_payment_enabled: bool = Field(..., alias="repeatPaymentEnabled")
    """Специальное поле"""

    _favorite_payment_enabled: bool = Field(
        ..., alias="favoritePaymentEnabled"
    )
    """Специальное поле"""

    _regular_payment_enabled: bool = Field(..., alias="regularPaymentEnabled")
    """Специальное поле"""

    def __str__(self) -> str:
        return f"""Статус транзакции: {self.status}
Дата проведения платежа: {self.date}
Сумма платежа: {self.sum.amount}
Комментарий: {self.comment if isinstance(self.comment, str) else '-'}
{"Кому" if self.type == "OUT" else "От кого"}: {self.to_account}
"""
